'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var JavascriptModulesPlugin = _interopDefault(require('webpack/lib/javascript/JavascriptModulesPlugin'));
var Template$2 = _interopDefault(require('webpack/lib/Template'));
var propertyAccess = _interopDefault(require('webpack/lib/util/propertyAccess'));
var validateOptions = _interopDefault(require('schema-utils'));
var ModuleDependency = _interopDefault(require('webpack/lib/dependencies/ModuleDependency'));
var webpackSources = require('webpack-sources');
var Dependency = _interopDefault(require('webpack/lib/Dependency'));
var ModuleFactory = _interopDefault(require('webpack/lib/ModuleFactory'));
var Module$2 = _interopDefault(require('webpack/lib/Module'));
var AsyncDependenciesBlock = _interopDefault(require('webpack/lib/AsyncDependenciesBlock'));
var RuntimeGlobals$2 = _interopDefault(require('webpack/lib/RuntimeGlobals'));

class ContainerExposedDependency extends ModuleDependency {
  constructor(name, request) {
    super(request);
    this._name = name;
  }

  get exposedName() {
    return this._name;
  }

  getResourceIdentifier() {
    return `exposed dependency ${this._name}`;
  }

}

class ContainerEntryDependency extends Dependency {
  constructor(dependencies, name) {
    super();
    this.exposedDependencies = dependencies;
    this.optional = true;
    this.loc = {
      name
    };
  }

}

const SOURCE_TYPES = new Set(['javascript']);
const RUNTIME_REQUIREMENTS = new Set([RuntimeGlobals$2.definePropertyGetters, RuntimeGlobals$2.exports, RuntimeGlobals$2.returnExportsFromRuntime]);
class ContainerEntryModule extends Module$2 {
  constructor(dependency) {
    super('javascript/dynamic', null);
    this.expose = dependency.exposedDependencies;
  }

  getSourceTypes() {
    return SOURCE_TYPES;
  }

  identifier() {
    return `container entry ${JSON.stringify(this.expose.map(item => item.exposedName))}`;
  }

  readableIdentifier() {
    return `container entry`;
  }

  needBuild(context, callback) {
    return callback(null, !this.buildMeta);
  }

  build(options, compilation, resolver, fs, callback) {
    this.buildMeta = {};
    this.buildInfo = {
      strict: true
    };
    this.clearDependenciesAndBlocks();

    for (const dep of this.expose) {
      const block = new AsyncDependenciesBlock(undefined, dep.loc, dep.userRequest);
      block.addDependency(dep);
      this.addBlock(block);
    }

    callback();
  }

  codeGeneration({
    moduleGraph,
    chunkGraph,
    runtimeTemplate
  }) {
    const sources = new Map();
    const runtimeRequirements = RUNTIME_REQUIREMENTS;
    const getters = [];

    for (const block of this.blocks) {
      const {
        dependencies: [dep]
      } = block;
      const name = dep.exposedName;
      const mod = moduleGraph.getModule(dep);
      const request = dep.userRequest;
      let str;

      if (!mod) {
        str = runtimeTemplate.throwMissingModuleErrorBlock({
          request: dep.userRequest
        });
      } else {
        str = `return ${runtimeTemplate.blockPromise({
          block,
          message: request,
          chunkGraph,
          runtimeRequirements
        })}.then(${runtimeTemplate.basicFunction('', `return ${runtimeTemplate.moduleRaw({
          module: mod,
          chunkGraph,
          request,
          weak: false,
          runtimeRequirements
        })}`)});`;
      }

      getters.push(`${Template$2.toNormalComment(`[${name}] => ${request}`)}"${name}": ${runtimeTemplate.basicFunction('', str)}`);
    }

    sources.set('javascript', new webpackSources.ConcatSource(`\n${runtimeTemplate.supportsConst() ? "const" : "var"} __MODULE_MAP__ = {${getters.join(',')}};`, `\n${runtimeTemplate.supportsConst() ? "const" : "var"} __GET_MODULE__ = ${runtimeTemplate.basicFunction(['module'], `return typeof __MODULE_MAP__[module] ==='function' ? __MODULE_MAP__[module].apply(null) : Promise.reject(new Error('Module ' + module + ' does not exist.'))`)};`, `\n\n${RuntimeGlobals$2.definePropertyGetters}(exports, {\n`, Template$2.indent([`get: ${runtimeTemplate.basicFunction('', 'return __GET_MODULE__')},`, `shared: ${runtimeTemplate.basicFunction(['module', 'getter'], '__webpack_require__.shared[module] = getter;')}`]), `});`));
    return {
      sources,
      runtimeRequirements
    };
  }

  size(type) {
    return 42;
  }

}

class ContainerEntryModuleFactory extends ModuleFactory {
  create({
    dependencies: [dependency]
  }, callback) {
    callback(null, {
      module: new ContainerEntryModule(dependency)
    });
  }

}

/*
	MIT License http://www.opensource.org/licenses/mit-license.php
	Author Tobias Koppers @sokra
*/

const {
  OriginalSource,
  RawSource
} = require('webpack-sources');

const Module = require('webpack/lib/Module');

const RuntimeGlobals = require('webpack/lib/RuntimeGlobals');

const Template = require('webpack/lib/Template');

const makeSerializable = require('webpack/lib/util/makeSerializable');
/** @typedef {import("webpack-sources").Source} Source */

/** @typedef {import("../declarations/WebpackOptions").WebpackOptions} WebpackOptions */

/** @typedef {import("./Chunk")} Chunk */

/** @typedef {import("./ChunkGraph")} ChunkGraph */

/** @typedef {import("./Compilation")} Compilation */

/** @typedef {import("./DependencyTemplates")} DependencyTemplates */

/** @typedef {import("./Module").CodeGenerationContext} CodeGenerationContext */

/** @typedef {import("./Module").CodeGenerationResult} CodeGenerationResult */

/** @typedef {import("./Module").LibIdentOptions} LibIdentOptions */

/** @typedef {import("./Module").NeedBuildContext} NeedBuildContext */

/** @typedef {import("./RequestShortener")} RequestShortener */

/** @typedef {import("./ResolverFactory").ResolverWithOptions} ResolverWithOptions */

/** @typedef {import("./RuntimeTemplate")} RuntimeTemplate */

/** @typedef {import("./WebpackError")} WebpackError */

/** @typedef {import("./util/Hash")} Hash */

/** @typedef {import("./util/fs").InputFileSystem} InputFileSystem */

/**
 * @param {string|string[]} variableName the variable name or path
 * @param {string} type the module system
 * @returns {string} the generated source
 */


const getSourceForGlobalVariableExternal = (variableName, type) => {
  if (!Array.isArray(variableName)) {
    // make it an array as the look up works the same basically
    variableName = [variableName];
  } // needed for e.g. window["some"]["thing"]


  const objectLookup = variableName.map(r => `[${JSON.stringify(r)}]`).join('');
  return `(function() { module.exports = ${type}${objectLookup}; }());`;
};
/**
 * @param {string|string[]} moduleAndSpecifiers the module request
 * @returns {string} the generated source
 */


const getSourceForCommonJsExternal = moduleAndSpecifiers => {
  if (!Array.isArray(moduleAndSpecifiers)) {
    return `module.exports = require(${JSON.stringify(moduleAndSpecifiers)});`;
  }

  const moduleName = moduleAndSpecifiers[0];
  const objectLookup = moduleAndSpecifiers.slice(1).map(r => `[${JSON.stringify(r)}]`).join('');
  return `module.exports = require(${JSON.stringify(moduleName)})${objectLookup};`;
};
/**
 * @param {string} variableName the variable name to check
 * @param {string} request the request path
 * @param {RuntimeTemplate} runtimeTemplate the runtime template
 * @returns {string} the generated source
 */


const checkExternalVariable = (variableName, request, runtimeTemplate) => {
  return `if(typeof ${variableName} === 'undefined') { ${runtimeTemplate.throwMissingModuleErrorBlock({
    request
  })} }\n`;
};
/**
 * @param {string|number} id the module id
 * @param {boolean} optional true, if the module is optional
 * @param {string|string[]} request the request path
 * @param {RuntimeTemplate} runtimeTemplate the runtime template
 * @returns {string} the generated source
 */


const getSourceForAmdOrUmdExternal = (id, optional, request, runtimeTemplate) => {
  const externalVariable = `__WEBPACK_EXTERNAL_MODULE_${Template.toIdentifier(`${id}`)}__`;
  const missingModuleError = optional ? checkExternalVariable(externalVariable, Array.isArray(request) ? request.join('.') : request, runtimeTemplate) : '';
  return `${missingModuleError}module.exports = ${externalVariable};`;
};
/**
 * @param {boolean} optional true, if the module is optional
 * @param {string|string[]} request the request path
 * @param {RuntimeTemplate} runtimeTemplate the runtime template
 * @returns {string} the generated source
 */


const getSourceForDefaultCase = (moduleId, optional, request, runtimeTemplate) => {
  if (!Array.isArray(request)) {
    // make it an array as the look up works the same basically
    request = [request];
  } // TODO: use this for error handling


  const missingModuleError = optional ? checkExternalVariable(request.join('.'), runtimeTemplate) : '';
  const requestScope = null; // refactor conditional into checkExternalVariable

  return Template.asString(['module.exports = ', `(typeof ${requestScope} !== 'undefined') ? __webpack_require__.shared('${request}') || Promise.resolve(__webpack_require__(${moduleId})) : `, `Promise.reject("Missing Shared Module: ${requestScope} cannot be found when trying to override ${request}"); `]);
};

const TYPES = new Set(['javascript']);
const RUNTIME_REQUIREMENTS$1 = new Set([RuntimeGlobals.module]);
class SharedModule extends Module {
  constructor(request, type, userRequest) {
    super('javascript/dynamic', null); // Info from Factory

    /** @type {string | string[] | Record<string, string | string[]>} */

    this.request = request;
    /** @type {string} */

    this.externalType = type;
    /** @type {string} */

    this.userRequest = userRequest;
  }
  /**
   * @returns {Set<string>} types availiable (do not mutate)
   */


  getSourceTypes() {
    return TYPES;
  }
  /**
   * @param {LibIdentOptions} options options
   * @returns {string | null} an identifier for library inclusion
   */


  libIdent(options) {
    return this.userRequest;
  }
  /**
   * @param {Chunk} chunk the chunk which condition should be checked
   * @param {Compilation} compilation the compilation
   * @returns {boolean} true, if the chunk is ok for the module
   */


  chunkCondition(chunk, {
    chunkGraph
  }) {
    return chunkGraph.getNumberOfEntryModules(chunk) > 0;
  }
  /**
   * @returns {string} a unique identifier of the module
   */


  identifier() {
    return 'external ' + JSON.stringify(this.request);
  }
  /**
   * @param {RequestShortener} requestShortener the request shortener
   * @returns {string} a user readable identifier of the module
   */


  readableIdentifier(requestShortener) {
    return 'external ' + JSON.stringify(this.request);
  }
  /**
   * @param {NeedBuildContext} context context info
   * @param {function(WebpackError=, boolean=): void} callback callback function, returns true, if the module needs a rebuild
   * @returns {void}
   */


  needBuild(context, callback) {
    return callback(null, !this.buildMeta);
  }
  /**
   * @param {WebpackOptions} options webpack options
   * @param {Compilation} compilation the compilation
   * @param {ResolverWithOptions} resolver the resolver
   * @param {InputFileSystem} fs the file system
   * @param {function(WebpackError=): void} callback callback function
   * @returns {void}
   */


  build(options, compilation, resolver, fs, callback) {
    this.buildMeta = {};
    this.buildInfo = {
      strict: true
    };
    callback();
  }

  getSourceString(runtimeTemplate, moduleGraph, chunkGraph) {
    const request = typeof this.request === 'object' && !Array.isArray(this.request) ? this.request[this.externalType] : this.request;
    console.log('externalType', this.externalType);

    switch (this.externalType) {
      case 'this':
      case 'window':
      case 'self':
        return getSourceForGlobalVariableExternal(request, this.externalType);

      case 'global':
        return getSourceForGlobalVariableExternal(request, runtimeTemplate.outputOptions.globalObject);

      case 'commonjs':
      case 'commonjs2':
        return getSourceForCommonJsExternal(request);

      case 'amd':
      case 'amd-require':
      case 'umd':
      case 'umd2':
      case 'system':
        return getSourceForAmdOrUmdExternal(chunkGraph.getModuleId(this), this.isOptional(moduleGraph), request, runtimeTemplate);

      default:
        return getSourceForDefaultCase(chunkGraph.getModuleId(this), this.isOptional(moduleGraph), request, runtimeTemplate);
    }
  }
  /**
   * @param {CodeGenerationContext} context context for code generation
   * @returns {CodeGenerationResult} result
   */


  codeGeneration({
    runtimeTemplate,
    moduleGraph,
    chunkGraph
  }) {
    const sourceString = this.getSourceString(runtimeTemplate, moduleGraph, chunkGraph);
    const sources = new Map();

    if (this.useSourceMap) {
      sources.set('javascript', new OriginalSource(sourceString, this.identifier()));
    } else {
      sources.set('javascript', new RawSource(sourceString));
    }

    return {
      sources,
      runtimeRequirements: RUNTIME_REQUIREMENTS$1
    };
  }
  /**
   * @param {string=} type the source type for which the size should be estimated
   * @returns {number} the estimated size of the module (must be non-zero)
   */


  size(type) {
    return 42;
  }
  /**
   * @param {Hash} hash the hash used to track dependencies
   * @param {ChunkGraph} chunkGraph the chunk graph
   * @returns {void}
   */


  updateHash(hash, chunkGraph) {// hash.update(this.externalType);
    // hash.update(JSON.stringify(this.request));
    // hash.update(
    // 	JSON.stringify(Boolean(this.isOptional(chunkGraph.moduleGraph))),
    // );
    // super.updateHash(hash, chunkGraph);
  }

  serialize(context) {
    const {
      write
    } = context;
    write(this.request);
    write(this.externalType);
    write(this.userRequest);
    super.serialize(context);
  }

  deserialize(context) {
    const {
      read
    } = context;
    this.request = read();
    this.externalType = read();
    this.userRequest = read();
    super.deserialize(context);
  }

}
makeSerializable(SharedModule, 'webpack/lib/SharedModule');

class ContainerPlugin {
  static get name() {
    return ContainerPlugin.constructor.name;
  }

  constructor(options) {
    var _options$name, _options$shared, _options$library, _options$libraryTarge, _options$filename, _options$expose;

    const name = (_options$name = options.name) !== null && _options$name !== void 0 ? _options$name : `remoteEntry`;
    validateOptions({
      type: 'object',
      properties: {
        shared: {
          type: 'object'
        },
        expose: {
          type: ['object', 'array']
        },
        name: {
          type: 'string',
          default: name
        },
        library: {
          type: 'string',
          default: name
        },
        libraryTarget: {
          type: 'string',
          default: 'global',
          enum: ['var', 'this', 'window', 'self', 'global', 'commonjs', 'commonjs2', 'amd', 'amd-require', 'umd', 'umd2', 'system']
        },
        filename: {
          anyOf: [{
            type: 'string'
          }, {
            instanceof: 'Function'
          }]
        }
      },
      additionalProperties: false
    }, options, {
      name: ContainerPlugin.name
    });
    this.options = {
      shared: (_options$shared = options.shared) !== null && _options$shared !== void 0 ? _options$shared : null,
      name,
      library: (_options$library = options.library) !== null && _options$library !== void 0 ? _options$library : name,
      libraryTarget: (_options$libraryTarge = options.libraryTarget) !== null && _options$libraryTarge !== void 0 ? _options$libraryTarge : 'global',
      filename: (_options$filename = options.filename) !== null && _options$filename !== void 0 ? _options$filename : undefined,
      // Undefined means, use the default behaviour
      expose: (_options$expose = options.expose) !== null && _options$expose !== void 0 ? _options$expose : {}
    };
  }

  apply(compiler) {
    var _compiler$name;

    if (compiler.options.optimization.runtimeChunk) {
      throw new Error('This plugin cannot integrate with RuntimeChunk plugin, please remote `optimization.runtimeChunk`.');
    }

    compiler.options.output.jsonpFunction = `${compiler.options.output.jsonpFunction}${(_compiler$name = compiler.name) !== null && _compiler$name !== void 0 ? _compiler$name : ''}${this.options.name}`;
    compiler.hooks.make.tapAsync(ContainerPlugin.name, (compilation, callback) => {
      let exposedMap = this.options.expose;

      if (Array.isArray(this.options.expose)) {
        exposedMap = {};

        for (const exp of this.options.expose) {
          // TODO: Check if this regex handles all cases
          exposedMap[exp.replace(/(^(?:[^\w])+)/, '')] = exp;
        }
      }

      compilation.addEntry(compilation.context, new ContainerEntryDependency(Object.entries(exposedMap).map(([name, request], idx) => {
        const dep = new ContainerExposedDependency(name, request);
        dep.loc = {
          name,
          index: idx
        };
        return dep;
      }), this.options.name), this.options.name, callback);
    }); // compiler.hooks.compile.tap(
    // 	ContainerPlugin.name,
    // 	({ normalModuleFactory }) => {
    // 		new SharedModuleFactoryPlugin(
    // 			compiler.options.libraryTarget,
    // 			this.options.shared,
    // 		).apply(normalModuleFactory);
    // 	},
    // );

    compiler.hooks.compilation.tap(ContainerPlugin.name, ({
      mainTemplate
    }) => {
      mainTemplate.hooks.requireExtensions.tap('URLImportPlugin', source => {
        return Template$2.asString([source, '__webpack_require__.shared = __webpack_require__']);
      });
    });
    compiler.hooks.thisCompilation.tap(ContainerPlugin.name, (compilation, {
      normalModuleFactory
    }) => {
      compilation.dependencyFactories.set(ContainerEntryDependency, new ContainerEntryModuleFactory());
      compilation.dependencyFactories.set(ContainerExposedDependency, normalModuleFactory);
      const renderHooks = JavascriptModulesPlugin.getCompilationHooks(compilation);
      renderHooks.render.tap(ContainerPlugin.name, (source, {
        chunk
      }) => {
        if (chunk.name === this.options.name) {
          const libName = Template$2.toIdentifier(compilation.getPath(this.options.library, {
            chunk
          }));

          switch (this.options.libraryTarget) {
            case 'var':
              {
                return new webpackSources.ConcatSource(`var ${libName} =`, source);
              }

            case 'this':
            case 'window':
            case 'self':
              return new webpackSources.ConcatSource(`${this.options.libraryTarget}${propertyAccess([libName])} =`, source);

            case 'global':
              return new webpackSources.ConcatSource(`${compiler.options.output.globalObject}${propertyAccess([libName])} =`, source);

            case 'commonjs':
            case 'commonjs2':
              {
                return new webpackSources.ConcatSource(`exports${propertyAccess([libName])} =`, source);
              }

            case 'amd': // TODO: Solve this?

            case 'amd-require': // TODO: Solve this?

            case 'umd': // TODO: Solve this?

            case 'umd2': // TODO: Solve this?

            case 'system': // TODO: Solve this?

            default:
              throw new Error(`${this.options.libraryTarget} is not a valid Library target`);
          }
        }
      });
      compilation.hooks.afterChunks.tap(ContainerPlugin.name, chunks => {
        for (const chunk of chunks) {
          if (chunk.name === this.options.name) {
            chunk.filenameTemplate = this.options.filename;
          }
        }
      });
    });
  }

}

const {
  OriginalSource: OriginalSource$1,
  RawSource: RawSource$1
} = require('webpack-sources');

const Module$1 = require('webpack/lib/Module');

const RuntimeGlobals$1 = require('webpack/lib/RuntimeGlobals');

const Template$1 = require('webpack/lib/Template');

const getSourceForGlobalVariableExternal$1 = (variableName, type, requestScope) => {
  if (!Array.isArray(variableName)) {
    // make it an array as the look up works the same basically
    variableName = [variableName];
  }

  const objectLookup = variableName.map(r => `${JSON.stringify(r)}`).join(''); // will output the following:
  // (function() {
  //   module.exports =
  //     typeof self["websiteTwo"] !== "undefined" ? self["websiteTwo"].get("Title") :
  //       Promise.reject("Missing Remote Runtime: self[\"websiteTwo\"] cannot be found when trying to import \"Title\"");
  // }());

  return Template$1.asString(['(function() {', // `console.log('shared',${type}["${requestScope}"])`,
  'module.exports =', `typeof ${type}["${requestScope}"] !== 'undefined' ? ${type}["${requestScope}"].get(${objectLookup}) : `, `Promise.reject('Missing Remote Runtime: ${type}["${requestScope}"] cannot be found when trying to import ${objectLookup}'); `, '}());']);
};
/**
 * @param {string|string[]} moduleAndSpecifiers the module request
 * @param {string|string[]} requestScope the module request namespace scope
 * @returns {string} the generated source
 */


const getSourceForCommonJsExternal$1 = (moduleAndSpecifiers, requestScope) => {
  if (!Array.isArray(moduleAndSpecifiers)) {
    // returns module.exports = require("websiteTwo").get("Title");
    return `module.exports = require(${JSON.stringify(requestScope)}).get(${JSON.stringify(moduleAndSpecifiers)});`;
  }

  const moduleName = moduleAndSpecifiers[0];
  const objectLookup = moduleAndSpecifiers.slice(1).map(r => `[${JSON.stringify(r)}]`).join(''); // returns module.exports = require("websiteTwo").get("Title")["default"];

  return `module.exports = require(${JSON.stringify(requestScope)}).get(${JSON.stringify(moduleName)})${objectLookup};`;
};
/**
 * @param {string} variableName the variable name to check
 * @param {string} request the request path
 * @param {RuntimeTemplate} runtimeTemplate the runtime template
 * @returns {string} the generated source
 */


const checkExternalVariable$1 = (variableName, request, runtimeTemplate) => {
  return `if(typeof ${variableName} === 'undefined') { ${runtimeTemplate.throwMissingModuleErrorBlock({
    request
  })} }\n`;
};
/**
 * @param {boolean} optional true, if the module is optional
 * @param {string|string[]} request the request path
 * @param {RuntimeTemplate} runtimeTemplate the runtime template
 * @returns {string} the generated source
 */


const getSourceForDefaultCase$1 = (optional, request, runtimeTemplate, requestScope) => {
  if (!Array.isArray(request)) {
    // make it an array as the look up works the same basically
    request = [request];
  } // TODO: use this for error handling


  const missingModuleError = optional ? checkExternalVariable$1(requestScope, request.join('.'), runtimeTemplate) : ''; // refactor conditional into checkExternalVariable

  return Template$1.asString(['module.exports = ', `typeof ${requestScope} !== 'undefined' ? ${requestScope}.get('${request}') : `, `Promise.reject("Missing Remote Runtime: ${requestScope} cannot be found when trying to import ${request}"); `]);
};

const TYPES$1 = new Set(['javascript']);
const RUNTIME_REQUIREMENTS$2 = new Set([RuntimeGlobals$1.module]);
class RemoteModule extends Module$1 {
  constructor(request, type, userRequest) {
    var _request$split, _request$split$shift, _request$split2;

    super('javascript/dynamic', null);
    this.requestScope = request === null || request === void 0 ? void 0 : (_request$split = request.split('/')) === null || _request$split === void 0 ? void 0 : (_request$split$shift = _request$split.shift) === null || _request$split$shift === void 0 ? void 0 : _request$split$shift.call(_request$split); // Info from Factory

    /** @type {string | string[] | Record<string, string | string[]>} */

    this.request = request === null || request === void 0 ? void 0 : (_request$split2 = request.split(`${this.requestScope}/`)) === null || _request$split2 === void 0 ? void 0 : _request$split2[1];
    /** @type {string} */

    this.remoteType = type;
    /** @type {string} */

    this.userRequest = userRequest;
  }
  /**
   * @returns {Set<string>} types availiable (do not mutate)
   */


  getSourceTypes() {
    return TYPES$1;
  }
  /**
   * @param {LibIdentOptions} options options
   * @returns {string | null} an identifier for library inclusion
   */


  libIdent(options) {
    return this.userRequest;
  }
  /**
   * @param {Chunk} chunk the chunk which condition should be checked
   * @param {Compilation} compilation the compilation
   * @returns {boolean} true, if the chunk is ok for the module
   */


  chunkCondition(chunk, {
    chunkGraph
  }) {
    return chunkGraph.getNumberOfEntryModules(chunk) > 0;
  }
  /**
   * @returns {string} a unique identifier of the module
   */


  identifier() {
    return `remote ${JSON.stringify(this.request)}`;
  }
  /**
   * @param {RequestShortener} requestShortener the request shortener
   * @returns {string} a user readable identifier of the module
   */


  readableIdentifier(requestShortener) {
    return `remote ${JSON.stringify(this.request)}`;
  }
  /**
   * @param {NeedBuildContext} context context info
   * @param {function(WebpackError=, boolean=): void} callback callback function, returns true, if the module needs a rebuild
   * @returns {void}
   */


  needBuild(context, callback) {
    return callback(null, !this.buildMeta);
  }
  /**
   * @param {WebpackOptions} options webpack options
   * @param {Compilation} compilation the compilation
   * @param {ResolverWithOptions} resolver the resolver
   * @param {InputFileSystem} fs the file system
   * @param {function(WebpackError=): void} callback callback function
   * @returns {void}
   */


  build(options, compilation, resolver, fs, callback) {
    this.buildMeta = {};
    this.buildInfo = {
      strict: true
    };
    callback();
  }

  getSourceString(runtimeTemplate, moduleGraph, chunkGraph) {
    const request = typeof this.request === 'object' && !Array.isArray(this.request) ? this.request[this.remoteType] : this.request;

    switch (this.remoteType) {
      case 'this':
      case 'window':
      case 'self':
        return getSourceForGlobalVariableExternal$1(request, this.remoteType, this.requestScope);

      case 'global':
        return getSourceForGlobalVariableExternal$1(request, runtimeTemplate.outputOptions.globalObject, this.requestScope);

      case 'commonjs':
      case 'commonjs2':
        return getSourceForCommonJsExternal$1(request, this.requestScope);

      case 'amd':
      case 'amd-require':
      case 'umd':
      case 'umd2':
      case 'system':
        throw new Error(`${this.remoteType} is not supported with ContainerReferencePlugin`);

      default:
        return getSourceForDefaultCase$1(this.isOptional(moduleGraph), request, runtimeTemplate, this.requestScope);
    }
  }
  /**
   * @param {CodeGenerationContext} context context for code generation
   * @returns {CodeGenerationResult} result
   */


  codeGeneration({
    runtimeTemplate,
    moduleGraph,
    chunkGraph
  }) {
    const sourceString = this.getSourceString(runtimeTemplate, moduleGraph, chunkGraph);
    const sources = new Map();

    if (this.useSourceMap) {
      sources.set('javascript', new OriginalSource$1(sourceString, this.identifier()));
    } else {
      sources.set('javascript', new RawSource$1(sourceString));
    }

    return {
      sources,
      runtimeRequirements: RUNTIME_REQUIREMENTS$2
    };
  }
  /**
   * @param {string=} type the source type for which the size should be estimated
   * @returns {number} the estimated size of the module (must be non-zero)
   */


  size(type) {
    return 42;
  }
  /**
   * @param {Hash} hash the hash used to track dependencies
   * @param {ChunkGraph} chunkGraph the chunk graph
   * @returns {void}
   */


  updateHash(hash, chunkGraph) {
    // hash.update(this.remoteType);
    hash.update(JSON.stringify(this.request)); // hash.update(
    //   JSON.stringify(Boolean(this.isOptional(chunkGraph.moduleGraph)))
    // );
    // super.updateHash(hash, chunkGraph);
  }

  serialize(context) {
    const {
      write
    } = context;
    write(this.request);
    write(this.remoteType);
    write(this.userRequest);
    super.serialize(context);
  }

  deserialize(context) {
    const {
      read
    } = context;
    this.request = read();
    this.remoteType = read();
    this.userRequest = read();
    super.deserialize(context);
  }

}

const UNSPECIFIED_EXTERNAL_TYPE_REGEXP = /^[a-z0-9]+ /;

class RemoteModuleFactoryPlugin {
  constructor(type, remotes) {
    this.remoteType = type;
    this.remotes = remotes;
  }

  apply(normalModuleFactory) {
    const globalType = this.remoteType;
    normalModuleFactory.hooks.factorize.tapAsync('RemoteModuleFactoryPlugin', (data, callback) => {
      const dependency = data.dependencies[0];

      const handleRemote = (value, type, callback) => {
        if (value === false) {
          // Not externals, fallback to original factory
          return callback();
        }
        /** @type {string} */


        let externalConfig;

        if (value === true) {
          externalConfig = dependency.request;
        } else {
          externalConfig = value;
        } // When no explicit type is specified, extract it from the externalConfig


        if (type === undefined && UNSPECIFIED_EXTERNAL_TYPE_REGEXP.test(externalConfig)) {
          const idx = externalConfig.indexOf(' ');
          type = externalConfig.substr(0, idx);
          externalConfig = externalConfig.substr(idx + 1);
        }

        callback(null, new RemoteModule(externalConfig, type || globalType, dependency.request));
      };

      const handleExternals = (remotes, callback) => {
        if (typeof remotes === 'string') {
          var _data$request, _data$request$split, _data$request$split$s, _this$remotes;

          const requestScope = data === null || data === void 0 ? void 0 : (_data$request = data.request) === null || _data$request === void 0 ? void 0 : (_data$request$split = _data$request.split('/')) === null || _data$request$split === void 0 ? void 0 : (_data$request$split$s = _data$request$split.shift) === null || _data$request$split$s === void 0 ? void 0 : _data$request$split$s.call(_data$request$split);

          if ((_this$remotes = this.remotes) === null || _this$remotes === void 0 ? void 0 : _this$remotes.includes(requestScope)) {
            return handleRemote(dependency.request, undefined, callback);
          }
        } else if (Array.isArray(remotes)) {
          let i = 0;

          const next = () => {
            let asyncFlag;

            const handleExternalsAndCallback = (err, module) => {
              if (err) return callback(err);

              if (!module) {
                if (asyncFlag) {
                  asyncFlag = false;
                  return;
                }

                return next();
              }

              callback(null, module);
            };

            do {
              asyncFlag = true;
              if (i >= remotes.length) return callback();
              handleExternals(remotes[i++], handleExternalsAndCallback);
            } while (!asyncFlag);

            asyncFlag = false;
          };

          next();
          return;
        } else if (typeof remotes === 'object' && Object.prototype.hasOwnProperty.call(remotes, dependency.request)) {
          return handleRemote(remotes[dependency.request], undefined, callback);
        }

        callback();
      };

      handleExternals(this.remotes, callback);
    });
  }

}

class ContainerReferencePlugin {
  static get name() {
    return ContainerReferencePlugin.constructor.name;
  }

  constructor(options) {
    var _options$remoteType, _options$remotes, _options$override;

    this.options = {
      remoteType: (_options$remoteType = options.remoteType) !== null && _options$remoteType !== void 0 ? _options$remoteType : 'global',
      remotes: (_options$remotes = options.remotes) !== null && _options$remotes !== void 0 ? _options$remotes : [],
      override: (_options$override = options.override) !== null && _options$override !== void 0 ? _options$override : {}
    }; // TODO: Apply some validation around what was passed in.
  }

  apply(compiler) {
    const {
      remotes,
      remoteType
    } = this.options;
    compiler.hooks.compile.tap(ContainerReferencePlugin.name, ({
      normalModuleFactory
    }) => {
      new RemoteModuleFactoryPlugin(remoteType, remotes).apply(normalModuleFactory);
    });
  }

}

exports.ContainerPlugin = ContainerPlugin;
exports.ContainerReferencePlugin = ContainerReferencePlugin;
